﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Threading;

namespace _3DTransform
{
    public partial class Form3DTransform : Form
    {
        #region Attributes
        static int dim = 3; //3D Graphics
        static int D = dim + 1;
        static double d = 60;   //side of the cube
        double[,] verts0 = { { d, d, d, 1 }, 
                            { d, d, -d, 1 }, 
                            { -d, d, -d, 1 }, 
                            { -d, d, d, 1 }, 
                            { d, -d, d, 1 }, 
                            { d, -d, -d, 1 },
                            { -d, -d, -d, 1 },
                            { -d, -d, d, 1 }
                          };    //cube vertices
        double[,] verts;
        double[,] verts2;
        double[,] verts3;
        
        int[,] edges = { { 0, 1 }, { 1, 2 }, {2, 3}, { 3, 0 },
                        { 4, 5 }, { 5, 6 }, { 6, 7 }, { 7, 4},
                        { 0, 4 }, { 1, 5 }, { 2, 6 }, { 3, 7},
                       };       //cube edges

        double[,] rtxMat = new double[D, D];
        double[,] rtyMat = new double[D, D];
        double[,] rtzMat = new double[D, D];
        double[,] trMat = new double[D, D];
        double[,] scMat = new double[D, D];
        double[,] rfxyMat = new double[D, D];
        double[,] rfyzMat = new double[D, D];
        double[,] rfzxMat = new double[D, D];
        double[,] invMat = new double[D, D];
        double[,] shrMat = new double[D, D];
        double[,] resultMat = new double[D, D];

        enum Transform { TRANSLATION, ROTATIONX, ROTATIONY, ROTATIONZ, SCALING, REFLECTIONXY, REFLECTIONYZ, REFLECTIONZX, INVERSIONORIG, SHEARING,TR, RT, SCALINGANIM };
        Transform transform = (Transform)(-1);
        double theta = 0;
        double s = 1;
        double ds = 0.1;        //change in scaling for animation
        int counter = 0;
        Pen red = new Pen(Color.Red, 2); //Color.FromArgb(250,50,0),2);
        Pen green = new Pen(Color.Green, 2); //Color.FromArgb(0, 150, 150),2);
        int width = 800;
        int height = 600;
        int hw;                     //half width
        int hh;                     //half height
        #endregion
        public Form3DTransform()
        {
            InitializeComponent();
            hw = width / 2;
            hh = height / 2;
            int numPts = verts0.Length / D;
            verts = new double[numPts, D]; 
            verts2 = new double[numPts, D];
            verts3 = new double[numPts, D];           
            this.BackColor = Color.Black;
            this.CenterToScreen();
        }
        /// <summary>
        /// set all the matrices to identity matrices
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Form3DTransform_Load(object sender, EventArgs e)
        {
            setToIdentityMat(rtxMat);
            setToIdentityMat(rtyMat);
            setToIdentityMat(rtzMat);
            setToIdentityMat(trMat);
            setToIdentityMat(scMat);
            setToIdentityMat(rfxyMat);
            setToIdentityMat(rfyzMat);
            setToIdentityMat(rfzxMat);
            setToIdentityMat(invMat);
            setToIdentityMat(shrMat);
            setToIdentityMat(resultMat);

            setRotationXMat(Math.PI / 10);      //the cube looks 3 dimensional
            setRotationYMat(Math.PI / 10);      //
            matXMat(rtxMat, rtyMat);
            matXPoints(resultMat, verts0, verts);
        }

        private void Form3DTransform_Paint(object sender, PaintEventArgs e)
        {
            drawShape(verts, edges, green);
            //translateRotate(200, 0, 0);
            //rotateTranslate(200, 0, 0);
            scaleAnimate();
            //shear(0.5, -2, 1.5);
        }

        private void Form3DTransform_Resize(object sender, EventArgs e)
        {
            width = ClientRectangle.Width;
            height = ClientRectangle.Height;
            hw = width / 2;
            hh = height / 2;
            this.Invalidate();
        }
        public void drawShape(double[,] v, int[,] e, Pen p)
        {
            Graphics grfx = this.CreateGraphics();
            int n = e.Length / 2;
            for (int i = 0; i < n; i++)
            {
                grfx.DrawLine(p, hw + (int)v[e[i, 0], 0], hh - (int)v[e[i, 0], 1],
                                  hw + (int)v[e[i, 1], 0], hh - (int)v[e[i, 1], 1]
                             );

            }
            //grfx.DrawEllipse(p, hw+(int)v[2, 0]-5, hh-(int)v[2, 1]-5, 10, 10);
        }
        /// <summary>
        /// matrix is multiplied to points of given shape
        /// </summary>
        /// <param name="mat">matrix</param>
        /// <param name="pts1">vertices mutiplied</param>
        /// <param name="pts2">new vertices</param>
        public void matXPoints(double[,] mat, double[,] pts1, double[,] pts2)
        {
            int numPts = pts1.Length / D;
            double temp = 0;
            for (int n = 0; n < numPts; n++)
            {
                for (int r = 0; r < D; r++)
                {
                    for (int c = 0; c < D; c++)
                    {
                        temp = temp + mat[r, c] * pts1[n, c];
                    }
                    pts2[n, r] = temp;
                    temp = 0;
                }
            }
        }
        public void setTranslationMat(double a, double b,double c)
        {
            trMat[0, 3] = a;
            trMat[1, 3] = b;
            trMat[2, 3] = c;
        }
        public void setRotationXMat(double theta)
        {
            rtxMat[1, 1] = Math.Cos(theta);
            rtxMat[1, 2] = -Math.Sin(theta);
            rtxMat[2, 1] = Math.Sin(theta);
            rtxMat[2, 2] = Math.Cos(theta);
        }
        public void setRotationYMat(double theta)
        {
            rtyMat[0, 0] = Math.Cos(theta);
            rtyMat[0, 2] = Math.Sin(theta);
            rtyMat[2, 0] = -Math.Sin(theta);
            rtyMat[2, 2] = Math.Cos(theta);
        }
        public void setRotationZMat(double theta)
        {
            rtzMat[0, 0] = Math.Cos(theta);
            rtzMat[0, 1] = -Math.Sin(theta);
            rtzMat[1, 0] = Math.Sin(theta);
            rtzMat[1, 1] = Math.Cos(theta);
        }
        public void setScalingMat(double s)
        {
            scMat[0, 0] = s;
            scMat[1, 1] = s;
            scMat[2, 2] = s;
        }
        public void setShearingMat(double a, double b, double c)
        {
            shrMat[0, 0] = a;
            shrMat[0, 1] = b;
            shrMat[1, 1] = b;
            shrMat[1, 2] = c;
        }
        /// <summary>
        /// rotation about X-Axis
        /// </summary>
        /// <param name="angle">the rotation angle</param>
        public void rotateX(double angle)
        {
            setRotationXMat(angle);
            matXPoints(rtxMat, verts, verts2);
            drawShape(verts2, edges, red);
        }

        public void rotateY(double angle)
        {
            setRotationYMat(angle);
            matXPoints(rtyMat, verts, verts2);
            drawShape(verts2, edges, red);
        }
        public void rotateZ(double angle)
        {
            setRotationZMat(angle);
            matXPoints(rtzMat, verts, verts2);
            drawShape(verts2, edges, red);
        }
        public void translate(double a, double b,double c)
        {
            setTranslationMat(a, b,c);
            matXPoints(trMat, verts, verts2);
            drawShape(verts2, edges, red);
        }
        public void shear(double a, double b, double c)
        {
            setShearingMat(a, b, c);
            matXPoints(shrMat, verts, verts2);
            drawShape(verts2, edges, red);
        }
        /// <summary>
        /// vertices are first translated then rotated
        /// </summary>
        /// <param name="a"> x translation</param>
        /// <param name="b"> y translation</param>
        /// <param name="c"> z translation</param>
        public void translateRotate(double a, double b, double c)
        {
            setTranslationMat(a, b,c);
            setRotationZMat(theta);
            matXMat(rtzMat, trMat);
            matXPoints(resultMat, verts, verts2);
            drawShape(verts2, edges, red);
            theta = theta + 0.02;
            Thread.Sleep(20);
            if (theta < 2 * Math.PI)
                this.Invalidate();
            else theta = 0;
        }
        /// <summary>
        /// vertices are first rotated then translated
        /// </summary>
        /// <param name="a">x translation</param>
        /// <param name="b">y translation</param>
        /// <param name="c">z translation</param>
        public void rotateTranslate(double a, double b,double c)
        {
            setTranslationMat(a, b,c);
            setRotationYMat(theta);
            matXMat(trMat, rtyMat);
            matXPoints(resultMat, verts, verts2);
            drawShape(verts2, edges, red);
            theta = theta + 0.02;
            Thread.Sleep(20);
            if (theta < 2 * Math.PI)
                this.Invalidate();
            else theta = 0;
        }
        
        public void scaleAnimate()
        {
            setScalingMat(s);
            matXPoints(scMat, verts, verts2);
            drawShape(verts2, edges, red);
            Thread.Sleep(20);
            s = s + ds;
            if (s > 4 || s < 0.1)
                ds = -ds;
            counter = counter + 1;
            if (counter < 200)
                this.Invalidate();
            else counter = 0;
        }
        /// <summary>
        /// matrix multiplication.mat1 is multiplied with mat2 
        /// </summary>
        /// <param name="mat1"></param>
        /// <param name="mat2"></param>
        public void matXMat(double[,] mat1, double[,] mat2)
        {
            double temp;
            for (int row = 0; row < D; row++)
            {
                for (int col = 0; col < D; col++)
                {
                    temp = 0;
                    for (int k = 0; k < D; k++)
                    {
                        temp = temp + mat1[row, k] * mat2[k, col];
                    }
                    resultMat[row, col] = temp;
                }
            }
        }

        public void setToIdentityMat(double[,] m)
        {
            for (int row = 0; row < D; row++)
            {
                for (int col = 0; col < D; col++)
                {
                    if (row == col)
                        m[row, col] = 1;
                    else
                        m[row, col] = 0;
                }
            }
        }
    }
}
